/*******************************************************************************

Copyright 2002 Pavel Kurnosoff, pavkur@hotbox.ru.
                                                 All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
4. Neither the names of the authors nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <smapi/msgapi.h>
#include <time.h>

//#define DEBUG_LOG
#define NEED_RECODE

#define VERSION "1.2"
#define PID "nnfido-" VERSION
#define MAXSTR 8192
#define LINELENGTH 79

char *monthStrs[]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
char *NONESTR="none";


#ifdef NEED_RECODE
char *alt2win="+++++++++++ ";
char *win2alt="++'+\"++++++'+++++''\"\"++--*+'++++ +I+++++E\"++*I+i****Ne\"jSsiH";
char *recode=NULL, *recodeBack=NULL;

#define RECODE(c) (c)>127?recode[(c)-128]:(c)
#define RECODEBACK(c) (c)>127?recodeBack[(c)-128]:(c)
#else
#define RECODE(c) c
#endif

void recodeStr(unsigned char *s)
{
	while(*s) {
#ifdef NEED_RECODE
		*s=RECODE(*s);
#endif	
		if(*s<' ') *s=' ';
		s++;
	}
}

void recodeStrBack(unsigned char *s)
{
#ifdef RECODEBACK	
	while(*s) { *s=RECODEBACK(*s);s++; }
#endif	
}

int parseFtnAddr(char *s, NETADDR *a)
{
	char *p=s, *pn;
	int n=-1,wn=0, pq=1;

	if(!s || !*s) return 0;
	
	else { a->zone=a->net=a->node=a->point=0; }
	pn=p;
	while(*p && pq) {
		if(isdigit((int)*p))
			wn=1;
		else
			switch(*p) {
			case ':':
				if(!wn || n>0) return 0;
				a->zone=(wn<2)?atoi(pn):-1;
				pn=p+1;wn=0;n=1;
				break;
			case '/':
				if(!wn || n>1) return 0;
				a->net=(wn<2)?atoi(pn):-1;
				n=2;pn=p+1;wn=0;
				break;
			case '.':
				if(n>2) return 0;
				if(wn) a->node=(wn<2)?atoi(pn):-1;
				n=3;pn=p+1;
				if(!wn && n<0) break;
				if(!wn) return 0;
				wn=0;
				break;
			case '\n':
			case '\r':
			case '\0':	
			case ' ':
			case '\t':
			case '@':
				pq=0;
				break; 
			default:
				return 0;
			}
		p++;
	}
	switch(n) {
	case 2:
		if(!wn) return 0;
		a->node=(wn<2)?atoi(pn):-1;
		break;
	case 3:
		if(!wn) return 0;
		a->point=(wn<2)?atoi(pn):-1;
		break;
	case -1:
		if(!wn) return 0;
		a->node=(wn<2)?atoi(pn):-1;
		break;
	default:
		return 0;
	} 
	return 1;
}


char *addrStr(char *from, NETADDR *a)
{
	static char str[256];
	char efrom[36], *p, *q;
	q=efrom;
	for(p=from;*p;p++)
		if((*p>='A' && *p<='Z') || (*p>='a' && *p<='z') || (*p=='_') || (*p=='-') || (*p=='.'))
			*q++=*p;else if(*p==' ') *q++='.';
	*q=0;
	q=str;
	q+=sprintf(q, "%s <%s@", from, efrom);
	if(a->point) q+=sprintf(q, "p%d.", a->point);
	q+=sprintf(q, "f%d.n%d.z%d.fidonet.org>",
			   a->node, a->net,a->zone);
	return str;
}

char *ftnAddrStr(NETADDR *a)
{
	static char str[20];
	sprintf(str, "%d:%d/%d", a->zone, a->net, a->node);
	if(a->point) sprintf(str+strlen(str), ".%d", a->point);
	return str;
}

char *msgidStr(char *msgid)
{
	static char str[256];
	char *p, *s=str+1;
	if(!msgid) return "";
	str[0]='<';
	for(p=msgid;*p && s-str<240;p++) {
		if(isalnum(*p)) *s++=*p;
		else s+=sprintf(s,"$%02x",*p);
	}
	strcpy(s,"@nnfido.exe>");
	return str;
}

char *getKludge(char *ctl, char *name)
{
	char *end; int len, len2;
	static char value[256];
	while(*ctl) {
		if(*ctl==1) ctl++;
		else return NULL;
		end=strchr(ctl, 1);
		if(!end) end=strchr(ctl, 0);
		len=strlen(name);
		if(!strncasecmp(ctl, name, len)) {
			len2=end-(ctl+len);
			if(len2>MAXSTR) len2=MAXSTR;
			memcpy(value, ctl+len, len2);
			value[len2]=0;
			return value;
		}
		ctl=end;
	}
	return NULL;
}

int countLines(char *txt)
{
	int n=0;
	while(*txt) {
		if(*txt=='\r') n++;
		txt++;
	}
	return n;
}

void chomp(char *s)
{
	char *e=strchr(s,0)-1; while(*e=='\r' || *e=='\n') *e--=0;
}

MSG *openArea(char *p, int write)
{
	int t;
	switch(p[0]) {
	case '$': t=MSGTYPE_SQUISH|MSGTYPE_ECHO;p++;break;
	case '#': t=MSGTYPE_JAM   |MSGTYPE_ECHO;p++;break;
	case '@': t=MSGTYPE_SDM   |MSGTYPE_ECHO;p++;break;
	default:
		t=MSGTYPE_SDM;
	}
	return MsgOpenArea(p, write?MSGAREA_CRIFNEC:MSGAREA_NORMAL, t);
}

char *getFrom(char *ctl, XMSG *msg, char *txt)
{
	NETADDR fromaddr;
	char *msgid, *p;
	char *from=NULL;

	if(!from) {
		msgid=getKludge(ctl, "MSGID: ");
		if(msgid) {
			p=strchr(msgid, ' ');
			if(p) {
				*p=0; if(parseFtnAddr(msgid, &fromaddr)) from=strdup(addrStr(msg->from, &fromaddr));	*p=' ';
			}
		}
	}
	if(!from) 
		if((p=strstr(txt, " * Origin:")) && (p==txt || *(p-1)=='\r'))
			if((p=strrchr(p, '(')))
				if(parseFtnAddr(p+1, &fromaddr)) from=strdup(addrStr(msg->from, &fromaddr));
				
	if(!from) from=strdup(addrStr(msg->from, &msg->orig));
	return from;
}

void listHeaders(char *str)
{
	MSG *mh;
	MSGH *msgh;
	int mn;
	unsigned char *ctl, *txt, *p;
	if(strlen(str)<6) return;
	if(!(p=strchr(str+5,' '))) return;
	*p++=0;
	p++;
	if ((mh=openArea(str+5, 0))==NULL) return;
	while (1) {
		XMSG msg;
		char *reply, *from;
		mn=atoi(p);
		if ((msgh=MsgOpenMsg(mh, MOPEN_READ, mn))) {
			ctl=malloc(MsgGetCtrlLen(msgh)+1);
			txt=malloc(MsgGetTextLen(msgh)+1);
			MsgReadMsg(msgh, &msg, 0, MsgGetTextLen(msgh), txt, MsgGetCtrlLen(msgh), ctl);

			msg.from[XMSG_FROM_SIZE-1]=msg.to[XMSG_TO_SIZE-1]=msg.subj[XMSG_SUBJ_SIZE-1]=0;
			recodeStr(msg.from);
			recodeStr(msg.to);
			recodeStr(msg.subj);
			reply=strdup(msgidStr(getKludge(ctl, "REPLY: ")));
			from=getFrom(ctl, &msg, txt);
			printf("%d\t%s\t%s\t%02d %s %02d %02d:%02d:%02d\t%s\t%s\t%ld\t%d\t\t\n", mn,
				   strlen(msg.subj) ? (char *)msg.subj : NONESTR, from,
				   msg.date_written.date.da, monthStrs[msg.date_written.date.mo-1], (msg.date_written.date.yr+80)%100,
				   msg.date_written.time.hh, msg.date_written.time.mm, msg.date_written.time.ss,
				   msgidStr(getKludge(ctl, "MSGID: ")), reply,
				   MsgGetTextLen(msgh), countLines(txt)
				);
			free(from);
			free(reply);
			free(txt);
			free(ctl);
			MsgCloseMsg(msgh);
		}
		if((p=strchr(p,' '))) p++;else break;
	}	
	MsgCloseArea(mh);
}

void getArticle (char *str)
{
	MSG *mh;
	MSGH *msgh;
	XMSG msg;
	int mn;
	unsigned char *ctl, *txt, *p, *pend;
	if(strlen(str)<6) return;
	if(!(p=strchr(str+5,' '))) return;
	*p++=0;

	if ((mh=openArea(str+5, 0))==NULL) return;
	mn=atoi(p);
	if ((msgh=MsgOpenMsg(mh, MOPEN_READ, mn))!=NULL) {
		char *from;
		
		ctl=malloc(MsgGetCtrlLen(msgh)+1);
		txt=malloc(MsgGetTextLen(msgh)+1);
		MsgReadMsg(msgh, &msg, 0, MsgGetTextLen(msgh), txt, MsgGetCtrlLen(msgh), ctl);
		
		msg.from[XMSG_FROM_SIZE-1]=msg.to[XMSG_TO_SIZE-1]=msg.subj[XMSG_SUBJ_SIZE-1]=0;
		recodeStr(msg.from);
		recodeStr(msg.to);
		recodeStr(msg.subj);
		
		from=getFrom(ctl, &msg, txt);


		
		printf("From: %s\n", from);	free(from);
		printf("Subject: %s\n", strlen(msg.subj) ? (char *)msg.subj : NONESTR);
		printf("Date: %02d %s %04d %02d:%02d:%02d +0000\n",
			   msg.date_written.date.da, monthStrs[msg.date_written.date.mo-1], msg.date_written.date.yr+1980,
			   msg.date_written.time.hh, msg.date_written.time.mm, msg.date_written.time.ss
			   );
		printf("Message-ID: %s\n", msgidStr(getKludge(ctl, "MSGID: ")));
		if((p=getKludge(ctl, "REPLY: "))) printf("References: %s\n", msgidStr(p));
		printf("X-Comment-To: %s\n", msg.to);
		printf("Lines: %d\n", countLines(txt));

		for(p=ctl;p<ctl+MsgGetCtrlLen(msgh);) {
			if(*p!=1) return; else p++;
			printf("X-Kludge: ");
			for(;p<ctl+MsgGetCtrlLen(msgh) && *p!=1;p++) putchar(RECODE(*p));
			putchar('\n');
		}

		pend=txt+MsgGetTextLen(msgh);
		for(p=txt;p<pend;) {
			int len;
			if(*p==1) {
				fputs("X-Kludge: ", stdout);
				*p++=0;
				while(*p!='\r'||p==pend) {
					putchar(RECODE(*p));
					*p++=0;
				}
				if(p==pend) break;
				*p++=0;
				putchar('\n');
			} else if(!strncmp(p, "SEEN-BY:", 8)) {
				fputs("X-Seen-By: ", stdout);
				for(len=0;len<9;len++) *p++=0;
				while(*p!='\r'||p==pend) {
					putchar(*p);
					*p++=0;
				}
				if(p==pend) break;
				*p++=0;
				putchar('\n');
			} else {
				unsigned char *wbeg=NULL;
				len=0;
				while(*p!='\r'||p==pend) {
					if(!*p) *p=' ';else *p=RECODE(*p);
					len++;
					if(len>LINELENGTH+1 && wbeg!=NULL) {
						*wbeg='\n';
						len=p-wbeg;
					}
					if(*p==' ') wbeg=p;
					p++;
				}
				if(p==pend) break;
				*p++='\n';
			}
		}
		putchar('\n');
		for(p=txt;p<pend;p++) if(*p) putchar(*p);

		if(txt) free(txt);
		if(ctl) free(ctl);
		MsgCloseMsg(msgh);
	}
	MsgCloseArea(mh);
}

void postArticle(char *str)
{
	int textLength;
	MSG *mh;
	MSGH *msgh;
	XMSG msg;
	NETADDR orig, dest;
	unsigned char *p, *text, *q, *ctl;
	char temp[MAXSTR+1], reply[MAXSTR+1];
	time_t theTime;struct tm *ts;
	
	if(strlen(str)<6) return;
	if(!(p=strchr(str+5,' '))) return;
	*p++=0;
	parseFtnAddr(p, &orig);	
 	if(!(p=strchr(p,' '))) return;
 	parseFtnAddr(p+1, &dest);
	if(!(p=strchr(p+1,' '))) return;
	textLength=atoi(p);

	if((mh=openArea(str+5, 1))==NULL) return;
	if((msgh=MsgOpenMsg(mh, MOPEN_CREATE, 0))!=NULL) {
		memset(&msg, 0, sizeof(msg));
		
		fgets(reply, MAXSTR, stdin);chomp(reply);
		
		fgets(temp, MAXSTR, stdin);chomp(temp);
		recodeStrBack(temp);
		strncpy(msg.from, temp, XMSG_FROM_SIZE);
		
		fgets(temp, MAXSTR, stdin);chomp(temp);
		recodeStrBack(temp);
		strncpy(msg.to, temp, XMSG_TO_SIZE);
		
		fgets(temp, MAXSTR, stdin);chomp(temp);		
		recodeStrBack(temp);
		strncpy(msg.subj, temp, XMSG_SUBJ_SIZE);

		memcpy(&msg.orig, &orig, sizeof(NETADDR));
		memcpy(&msg.dest, &dest, sizeof(NETADDR));
		
		text=malloc(textLength+1);
		p=text;
		while(p<text+textLength) {
			int c=getchar();
			*p++=c=='\n'?'\r':c;
		}
		*p=0;
		recodeStrBack(text);

		theTime=time(NULL);ts=localtime(&theTime);
		
		msg.date_written.date.da=ts->tm_mday;
		msg.date_written.date.mo=ts->tm_mon+1;
		msg.date_written.date.yr=ts->tm_year+1900-1980;
		msg.date_written.time.hh=ts->tm_hour;
		msg.date_written.time.mm=ts->tm_min;
		msg.date_written.time.ss=ts->tm_sec;

		strftime(msg.__ftsc_date, 20, "%d %b %y %H:%M:%S", ts);

		msg.date_arrived=msg.date_written;
		
		msg.attr=MSGLOCAL;

		msg.replyto=0;
		memset(&msg.replies, 0, sizeof(msg.replies)); 
		
		ctl=malloc(MAXSTR);q=ctl;
		q+=sprintf(q, "\001PID: " PID "\001MSGID: %s %08lx",
				   ftnAddrStr(&msg.orig), theTime);
		if(strlen(reply)) {
			q+=sprintf(q, "\001REPLY: %s", reply);
		}
		MsgWriteMsg(msgh, 0, &msg, text, p-text, p-text, q-ctl, ctl);
		MsgCloseMsg(msgh);
		free(text);
	}
	MsgCloseArea(mh);
}

void groupInfo(char *str)
{
	MSG *mh;
	char *p;
	if(!(p=strchr(str+5, ' '))) return;
	*p++=0;
	if ((mh=openArea(str+5, 0))==NULL) return;
	printf("%s %ld %ld y\n", p, MsgGetHighMsg(mh), 1L);
	MsgCloseArea(mh);
}


int main(int argc, char *argv[])
{
	
	char str[MAXSTR+1];
	struct _minf minf;

#ifdef NEED_RECODE
	recode=alt2win;	recodeBack=win2alt;
#endif

	minf.req_version=0;
	minf.def_zone=2;	
	MsgOpenApi(&minf);
	while(fgets(str, MAXSTR, stdin)) {
		if(str[0]) str[strlen(str)-1]=0;
#ifdef DEBUG_LOG
		{
			FILE *f=fopen("c:/nnfido.log","at");
			fprintf(f,"==>%s<==\n",str);			
			fclose(f);
		}
#endif		
		if (!strncmp(str, "EXIT", 4)) return 0;
		if (!strncmp(str, "POST", 4))
			postArticle(str);
		else if (!strncmp(str, "STAT", 4)) groupInfo(str);
		else {
			if (!strncmp(str, "HEAD", 4)) listHeaders(str);
			else if (!strncmp(str, "RETR", 4)) getArticle(str);
			putchar(0);
			fflush(stdout);
		}
	}
	MsgCloseApi();
	return 0;
}

